__version__ = "0.4.3"

from joker.clients.cas import ContentAddressedStorageClient, CascadisClient
from joker.clients.monolog import MonologInterface
from joker.clients.printable import PrintableClient

CASClient = ContentAddressedStorageClient
MonologClient = MonologInterface

if __name__ == "__main__":
    print(__version__)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import logging
import mimetypes
import os.path
import smtplib
import socket
import warnings
from dataclasses import dataclass
from email import encoders
from email.message import EmailMessage
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from functools import cached_property
from typing import TypedDict, Union, List

_logger = logging.getLogger(__name__)


class ExtendedEmailMessage(EmailMessage):
    @staticmethod
    def _fmt_addrs(addrs: Union[List[str], str], sep=", "):
        if isinstance(addrs, str):
            return addrs
        return sep.join(addrs)

    def set_subject_and_addrs(
        self,
        subject: str,
        from_addr: str,
        to_addrs: Union[List[str], str],
        cc_addrs: Union[List[str], str] = None,
    ):
        self["Subject"] = subject
        self["From"] = from_addr
        self["To"] = self._fmt_addrs(to_addrs)
        if cc_addrs:
            self["Cc"] = self._fmt_addrs(cc_addrs)

    def add_attachment_from_local_file(self, path: str):
        mime = mimetypes.guess_type(path)[0]
        maintype, subtype = mime.split("/", 1)
        filename = os.path.split(path)[1]
        with open(path, "rb") as fin:
            content = fin.read()
            self.add_attachment(
                content,
                filename=filename,
                maintype=maintype,
                subtype=subtype,
            )


# deprecated; will be removed at ver 0.5.0
class EmailInterface:
    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        password: str,
        from_addr: str,
        fake=False,
    ):
        """
        Args:
            host:
            port:
            username:
            password:
            from_addr:
            fake: if true, messages will NOT be send actually
        """
        warnings.warn(
            "EmailInterface is deprecated and will be removed at ver 0.5.0",
            DeprecationWarning,
        )
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.from_addr = from_addr
        self.fake = fake

    def _get_smtp(self):
        smtp = smtplib.SMTP(self.host, self.port)
        smtp.login(self.username, self.password)
        return smtp

    def _send(self, from_addr, to_addrs, msg):
        smtp = self._get_smtp()
        try:
            resp = smtp.sendmail(from_addr, to_addrs, msg)
            _logger.info("smtp.sendmail() => %s", resp)
        finally:
            smtp.quit()
            _logger.info("smtp.quit()")
        return resp

    def send(
        self,
        subject: str,
        content: str,
        attachments: list,
        from_addr: Union[str, None],
        to_addrs: Union[List[str], str],
        cc_addrs: Union[List[str], str] = None,
        content_subtype="html",
    ):
        """
        Args:
            subject:
            content:
            attachments
            from_addr: override default from_addr
            to_addrs:
            cc_addrs:
            content_subtype:
        """
        from_addr = from_addr or self.from_addr
        mail = ExtendedEmailMessage()
        mail.set_subject_and_addrs(
            subject,
            from_addr,
            to_addrs,
            cc_addrs,
        )
        mail.set_content(content, subtype=content_subtype)
        for att in attachments:
            mail.add_attachment_from_local_file(att)

        msg = mail.as_string()
        _logger.debug("EmailMessage().as_string()[:100] => %s", msg[:100])
        _logger.info("EmailMessage().fake => %s", self.fake)
        if not self.fake:
            return self._send(from_addr, to_addrs, msg)


class EmailAccount(TypedDict):
    user: str
    password: str


class EmailAgentConfDict(TypedDict):
    host: str
    port: int
    accounts: list[EmailAccount]
    interceptors: Union[list[str], None]


@dataclass
class EmailAgent:
    host: str
    port: int
    accounts: list[EmailAccount]
    interceptors: Union[list[str], None] = None

    @cached_property
    def _local_hostname(self) -> str:
        return socket.getfqdn()

    def _look_for_account(self, user: str) -> EmailAccount:
        for account in self.accounts:
            if account["user"] == user:
                return account
        raise LookupError(f"email account {user} not found")

    def login(self, user: str):
        smtp = smtplib.SMTP(
            self.host,
            self.port,
            local_hostname=self._local_hostname,
        )
        account = self._look_for_account(user)
        smtp.login(**account)
        return smtp

    def _prep_msg(
        self,
        from_addr: str,
        to_addrs: list,
        subject: str,
    ) -> MIMEMultipart:
        """Forge 'To' and 'Subject' if self.interceptors is configured."""
        msg = MIMEMultipart()
        msg["From"] = from_addr
        if self.interceptors:
            to = ", ".join(self.interceptors)
            msg["To"] = to
            msg["Subject"] = f"Intercepted: {subject}"
            text = f"This is an intercepted mail sending to {to}."
            msg.attach(MIMEText(text, "plain"))
        else:
            msg["To"] = ", ".join(to_addrs)
            msg["Subject"] = subject
        return msg

    def send(
        self,
        from_addr: str,
        to_addrs: list,
        subject: str,
        body: str | MIMEBase,
        attachments: list[MIMEBase] = None,
    ):
        """
        Note:
            SMTP.send_message() is a convenience method for calling SMTP.sendmail()
            with the message represented by an email.message.Message object.
            https://docs.python.org/3/library/smtplib.html#smtplib.SMTP.send_message
        """
        msg = self._prep_msg(from_addr, to_addrs, subject)
        if isinstance(body, str):
            body = MIMEText(body, "plain")
        msg.attach(body)
        attachments = attachments or []
        for attachment in attachments:
            msg.attach(attachment)
        with self.login(from_addr) as smtp:
            # If from_addr is None or to_addrs is None, send_message() fills
            # those arguments with addresses extracted from the headers of
            # msg as specified in RFC 5322.
            smtp.send_message(msg)
            # alternatively, one can use:
            # smtp.sendmail(from_addr, to_addrs, msg.as_string())

    @staticmethod
    def create_attachment(content: bytes, filename: str) -> MIMEBase:
        maintype, subtype = mimetypes.guess_type(filename)[0].split("/")
        attachment = MIMEBase(maintype, subtype)
        attachment.set_payload(content)
        # why encode in base64?
        # https://stackoverflow.com/a/76324353/2925169
        encoders.encode_base64(attachment)
        attachment.add_header(
            "Content-Disposition",
            "attachment",
            filename=("utf-8", "", filename),
        )
        return attachment



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import logging
from urllib.parse import urljoin

from joker.clients.utils import decode_response, _HTTPClient

_logger = logging.getLogger(__name__)


class MonologInterface(_HTTPClient):
    def fmt_url(self, channel: str):
        return urljoin(self.url, f"api/v1/{channel}")

    def push(self, channel: str, data: list):
        """Push data to the monolog server"""
        url = self.fmt_url(channel)
        _logger.info(
            "pushing %r records to monolog channel %r, %r", len(data), channel, url
        )
        resp = self.session.post(url, json=data)
        decode_response(resp)

    def pull(
        self, channel: str, since: str, limit: int = 1000, timeout: int = None
    ) -> list[dict]:
        """Pull data from the monolog server"""
        url = self.fmt_url(channel)
        _logger.info(
            "pulling %r records from monolog channel %r since %r, %r",
            limit,
            channel,
            since,
            url,
        )
        params = {"since": since, "limit": limit, "timeout": timeout}
        params = {k: v for k, v in params.items() if v is not None}
        resp = self.session.get(url, params=params)
        return decode_response(resp)

    add = push
    fetch = pull



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

from urllib.parse import urljoin

import requests


class _NamtsoClient:
    def __init__(self, base_url: str):
        self.base_url = base_url


class NamtsoResolverClient(_NamtsoClient):
    def acquire(self, kind: str):
        url = urljoin(self.base_url, "acquire")
        resp = requests.post(url, json={"kind": kind})
        return resp.json()["data"]

    def release(self, vtid: str, output: dict):
        url = urljoin(self.base_url, "release")
        resp = requests.post(url, json={"vtid": vtid, "output": output})
        return resp.json()["data"]


class NamtsoInquirerClient(_NamtsoClient):
    def publish(self, kind: str, input_: dict) -> str:
        url = urljoin(self.base_url, "publish")
        resp = requests.post(url, json={"kind": kind, "input": input_})
        return resp.json()["data"]

    def inquire(self, vtid: str, timeout: int | float = 0):
        url = urljoin(self.base_url, "inquire")
        resp = requests.post(url, json={"vtid": vtid, "timeout": timeout})
        return resp.json()["data"]



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import logging
from dataclasses import dataclass
from functools import cached_property
from typing import TypedDict
from urllib.parse import urljoin

import requests
from volkanic.errors import TechnicalError

from joker.clients import utils

_logger = logging.getLogger(__name__)


class PrintableTaskDict(TypedDict):
    tpl_path: str
    ctxid: str
    html_url: str
    pdf_url: str


@dataclass
class PrintableTask:
    client: PrintableClient
    tpl_path: str
    ctxid: str

    def _fmt_url(self, base: str, path: str) -> str:
        url = urljoin(base, path)
        return url + f"?ctxid={self.ctxid}"

    @property
    def inner_html_url(self) -> str:
        return self._fmt_url(self.client.inner_url, self.tpl_path)

    @property
    def inner_pdf_url(self) -> str:
        return self._fmt_url(self.client.inner_url, f"{self.tpl_path}.pdf")

    @property
    def outer_html_url(self) -> str:
        return self._fmt_url(self.client.outer_url, self.tpl_path)

    @property
    def outer_pdf_url(self) -> str:
        return self._fmt_url(self.client.outer_url, f"{self.tpl_path}.pdf")

    def to_dict(self) -> PrintableTaskDict:
        return {
            "tpl_path": self.tpl_path,
            "ctxid": self.ctxid,
            "html_url": self.outer_html_url,
            "pdf_url": self.outer_pdf_url,
        }

    def obtain_html(self) -> str:
        return self.client.session.get(self.inner_html_url).text

    def obtain_pdf(self) -> bytes:
        return self.client.session.get(self.inner_pdf_url).content


@dataclass
class PrintableClient:
    inner_url: str
    outer_url: str = None

    @classmethod
    def from_cfg(cls, cfg: str | dict):
        if isinstance(cfg, str):
            return cls(cfg)
        return cls(**cfg)

    @cached_property
    def session(self):
        return requests.Session()

    def __post_init__(self):
        utils.ensure_url_root(self.inner_url)
        if self.outer_url is None:
            self.outer_url = self.inner_url
        c = self.__class__.__name__
        _logger.info("new %s instance, %r", c, self.inner_url)

    @staticmethod
    def _post_as_json(url, data, allow_redirects=False) -> requests.Response:
        resp = utils.post_as_json(url, data, allow_redirects=allow_redirects)
        status = resp.status_code
        if status >= 400:
            raise TechnicalError(f"got response status code {status}")
        return resp

    def begin(self, tpl_path: str, data: dict) -> PrintableTask:
        assert tpl_path.endswith(".html")
        url = urljoin(self.inner_url, tpl_path)
        url += ".pdf"
        _logger.info("begin context with url: %r", url)
        resp = self._post_as_json(url, data, allow_redirects=False)
        try:
            ctxid = utils.parse_url_qsd(resp.headers["Location"])["ctxid"]
        except KeyError:
            raise RuntimeError(f"failed to render {url!r}")
        return PrintableTask(self, tpl_path, ctxid)

    def _generate(self, tpl_path: str, data: dict) -> tuple[bytes, str]:
        url = urljoin(self.inner_url, tpl_path)
        _logger.info("initial url: %r", url)
        resp = self._post_as_json(url, data, allow_redirects=True)
        _logger.info("redirected url: %r", resp.url)
        _logger.info(
            "content: %s bytes, %r",
            len(resp.content),
            resp.content[:100],
        )
        if not utils.check_pdf_validity(resp.content):
            raise RuntimeError("corrupted PDF file")
        return resp.content, resp.url

    def render_pdf(self, tpl_path: str, data: dict) -> bytes:
        assert tpl_path.endswith(".pdf")
        return self._generate(tpl_path, data)[0]

    def render_html(self, tpl_path: str, data: dict) -> str:
        assert tpl_path.endswith(".html")
        url = urljoin(self.inner_url, tpl_path)
        return self._post_as_json(url, data).text

    @property
    def base_url(self):
        """for backward-compatibility"""
        return self.inner_url



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import json
import logging
import os
import shlex
import typing
import urllib.parse
from dataclasses import dataclass
from functools import cached_property
from json import JSONDecodeError
from typing import Callable

import requests
from volkanic.errors import TechnicalError
from volkanic.introspect import razor
from volkanic.utils import json_default

_logger = logging.getLogger(__name__)
Pathlike = typing.Union[str, os.PathLike]


class ResponseDict(dict):
    @property
    def code(self) -> int:
        return self.get("code", 0)

    @property
    def data(self):
        return self.get("data")

    @property
    def message(self):
        return self.get("message")


def dump_json_request_to_curl(method: str, url: str, data=None, aslist=False):
    method = method.upper()
    if method == "GET":
        parts = ["curl", url]
    else:
        parts = [
            "curl",
            "-X",
            method,
            url,
            "-H",
            "Content-Type: application/json",
            "-d",
            json.dumps(razor(data), ensure_ascii=False),
        ]
    if aslist:
        return parts
    parts = [shlex.quote(s) for s in parts]
    return " ".join(parts)


def _log_bad_response(resp: requests.Response):
    _logger.error("bad response: %s %r", resp.status_code, resp.content[:1000])


def _decode_response(resp: requests.Response):
    status = resp.status_code
    if status >= 400:
        raise TechnicalError(f"got response status code {status}")
    try:
        rd = ResponseDict(resp.json())
    except JSONDecodeError:
        raise TechnicalError("cannot decode json")
    if rd.code != 0:
        raise TechnicalError(f"error response ({rd.code})")
    if rd.message:
        _logger.info(rd.message)
    return rd.data


def decode_response(resp: requests.Response):
    try:
        return _decode_response(resp)
    except TechnicalError:
        _log_bad_response(resp)
        raise


def parse_url_qsd(url: str) -> dict:
    """
    >>> parse_url_qsd('https://example.com/?q=1&q=2')
    {'q': '2'}
    >>> parse_url_qsd('https://example.com/')
    {}
    """
    query = urllib.parse.urlparse(url).query
    return dict(urllib.parse.parse_qsl(query))


def ensure_url_root(url: str) -> None:
    path = urllib.parse.urlparse(url).path
    if path == "" or path.endswith("/"):
        return
    raise ValueError('service url path must end with "/"')


def post_as_json(url: str, data: dict, func: Callable = requests.post, **kwargs):
    """
    Exists because by calling requests.post(url, json=data)
    you have nowhere to pass a parameter like default=str
    """
    headers = kwargs.setdefault("headers", {})
    headers.update({"Content-Type": "application/json"})
    payload = json.dumps(data, default=json_default)
    return func(url, data=payload, **kwargs)


@dataclass
class _HTTPClient:
    url: str

    def __post_init__(self):
        ensure_url_root(self.url)
        c = self.__class__.__name__
        _logger.info("new %s instance, %r", c, self.url)

    @cached_property
    def session(self):
        return requests.Session()

    def _post_as_json(self, url: str, data: dict, **kwargs):
        """
        Exists because by calling requests.post(url, json=data)
        you have nowhere to pass a parameter like default=str
        """
        return post_as_json(url, data, func=self.session.post, **kwargs)

    @property
    def base_url(self):
        """for backward-compatibility"""
        return self.url


_BaseHTTPClient = _HTTPClient


def check_pdf_validity(pdf_content: bytes) -> bool:
    if not pdf_content.startswith(b"%PDF-"):
        return False
    if len(pdf_content) < 1000:
        return False
    if b"%%EOF" not in pdf_content[-1024:]:
        return False
    return True



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import json
import logging
import traceback
from typing import Union
from urllib.parse import urljoin

import requests

from joker.clients.utils import (
    ResponseDict,
    dump_json_request_to_curl,
    _log_bad_response,
    _decode_response,
    decode_response,
    _BaseHTTPClient,
)

_logger = logging.getLogger(__name__)

_compat_names = [
    decode_response,
    _decode_response,
    _log_bad_response,
    dump_json_request_to_curl,
    ResponseDict,
    _BaseHTTPClient,
]
_JSON_TYPE = Union[dict, list, None, int, float, bool]


class HTTPClient(_BaseHTTPClient):
    timeout = 30

    def get_url(self, path: str):
        return urljoin(self.base_url, path)

    @staticmethod
    def _load_json(resp: requests.Response) -> _JSON_TYPE:
        err = _logger.error
        if resp.status_code != 200:
            err("status_code %s from %s", resp.status_code, resp.url)
            return
        try:
            data = json.loads(resp.text)
        except json.JSONDecodeError:
            traceback.print_exc()
            err("bad json from %s", resp.url)
            return
        if not isinstance(data, dict):
            err("wrong response type (%s) from %s", type(data), resp.url)
        code = data.get("code")
        if data.get("code"):
            err("non-zero code (%s) from %s", code, resp.url)
            return
        return data

    @staticmethod
    def log_request_as_curl(method: str, url: str, data=None, level=logging.DEBUG):
        if not _logger.isEnabledFor(level):
            return
        cmd = dump_json_request_to_curl(method, url, data=data)
        _logger.log(level, cmd)

    @staticmethod
    def log_resp_content(resp: requests.Response, level=logging.ERROR):
        if not _logger.isEnabledFor(level):
            return
        try:
            _logger.log(level, resp.text[:255])
        except Exception:
            _logger.log(level, resp.content[:255])
            traceback.print_exc()

    def json_request(self, method, path, data=None, **kwargs) -> _JSON_TYPE:
        url = self.get_url(path)
        method = method.upper()
        kwargs.setdefault("timeout", self.timeout)
        if data is not None:
            kwargs["data"] = json.dumps(data, default=str)
            kwargs["headers"] = {"Content-type": "application/json"}
        resp = requests.request(method, url, **kwargs)
        resp_data = self._load_json(resp)
        if resp_data is None:
            self.log_resp_content(resp)
            self.log_request_as_curl(method, url, data)
            return
        self.log_request_as_curl(method, url, data, level=logging.DEBUG)
        return resp_data

    def json_get(self, path, **kwargs) -> _JSON_TYPE:
        assert "data" not in kwargs
        return self.json_request("GET", path, **kwargs)

    def json_post(self, path, data=None, **kwargs) -> _JSON_TYPE:
        assert "data" not in kwargs
        return self.json_request("POST", path, data=data, **kwargs)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import os
import typing
import zipfile
import zlib
from dataclasses import dataclass
from functools import cached_property
from urllib.parse import urljoin

import requests
from volkanic.errors import TechnicalError

from joker.clients import utils
from joker.clients.utils import Pathlike


class MemberFile(typing.TypedDict):
    cid: str
    filename: str


@dataclass
class _CASClientBase:
    inner_url: str
    outer_url: str = None

    def __post_init__(self):
        utils.ensure_url_root(self.inner_url)
        if self.outer_url is None:
            self.outer_url = self.inner_url
        else:
            utils.ensure_url_root(self.outer_url)

    @cached_property
    def session(self):
        # note: requests.session() is deprecated; use cap S
        return requests.Session()

    def save(self, content: bytes) -> str:
        raise NotImplementedError

    def load(self, cid: str) -> None | bytes:
        raise NotImplementedError

    def upload(self, path: Pathlike):
        content = open(path, "rb").read()
        return self.save(content)

    def download(self, cid: str, path: Pathlike):
        rand_hex = os.urandom(8).hex()
        tmp_path = f"{path}.{rand_hex}.tmp"
        with open(tmp_path, "wb") as fout:
            fout.write(self.load(cid))
        os.rename(tmp_path, path)


@dataclass
class ContentAddressedStorageClient(_CASClientBase):
    inner_url: str
    outer_url: str = None

    def save(self, content: bytes) -> str:
        url = urljoin(self.inner_url, "files")
        resp = self.session.post(url, files={"file": content})
        return resp.json()["data"]

    def load(self, cid: str) -> None | bytes:
        url = urljoin(self.inner_url, f"files/{cid}")
        resp = requests.get(url)
        if resp.status_code == 200:
            return resp.content
        if resp.status_code == 404:
            return
        raise TechnicalError(f"got response status code {resp.status_code}")

    def load_text(self, cid: str) -> None | str:
        content = self.load(cid)
        if not content:
            return
        return zlib.decompress(content, wbits=31).decode("utf-8")

    def save_text(self, text: str) -> str:
        content = zlib.compress(text.encode("utf-8"), wbits=31)
        return self.save(content)

    def get_outer_url(self, cid: str, filename: str):
        if filename.startswith("."):
            url = f"files/{cid}{filename}"
        else:
            url = f"files/{cid}?filename={filename}"
        return urljoin(self.outer_url, url)

    def create_archive(self, path: Pathlike, memberfiles: list[MemberFile]):
        with zipfile.ZipFile(path, "w") as zipf:
            for m in memberfiles:
                content = self.load(m["cid"])
                with zipf.open(m["filename"], "w") as fout:
                    fout.write(content)


@dataclass
class CascadisClient(_CASClientBase):
    inner_url: str
    outer_url: str = None

    def save(self, content: bytes) -> str:
        url = urljoin(self.inner_url, "cascadis/files")
        resp = self.session.post(url, files={"file": content})
        return resp.json()["data"]

    def load(self, cid: str) -> None | bytes:
        url = urljoin(self.inner_url, f"cascadis/content/{cid}")
        resp = requests.get(url)
        if resp.status_code == 200:
            return resp.content
        if resp.status_code == 404:
            return
        raise TechnicalError(f"got response status code {resp.status_code}")

    def fmt_outer_url(self, cid: str, filename: str, short=False):
        prefix = "cas/c" if short else "cascadis/content"
        if filename.startswith("."):
            url = f"{prefix}/{cid}{filename}"
        else:
            url = f"{prefix}/{cid}?filename={filename}"
        return urljoin(self.outer_url, url)
